﻿using Butterfly.Core;
using Butterfly.Exceptions;
using Butterfly.HabboHotel;
using Butterfly.HabboHotel.Pets;
using Butterfly.HabboHotel.Users.ChatMessageHandling.Invoking;
using Butterfly.Messages.ClientMessages;
using Butterfly.Messages.StaticMessageHandlers;
using Butterfly.Net;
using Butterfly.ServerManager;
using Database_Manager.Database;
using Database_Manager.Database.Session_Details.Interfaces;
using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Runtime.Serialization.Json;

namespace Butterfly
{
    static class ButterflyEnvironment
    {
        public static Encoding DefaultEncoding { get; private set; }
        public static ConnectionHandeling ConnectionManager { get; private set; }
        public static Game Game { get; private set; }
        public static DatabaseManager DatabaseManager { get; private set; }
        public static CommandInvoker CommandInvoker { get; private set; }
        public static ButterflyConfiguration Configuration { get; private set; }
        
        internal const string VersionString = "Butterfly Emulator Version r107 codename buttercraft";
        internal static bool displayPacketActivity;
        private static MusSocket mus;

        internal static DateTime ServerStarted { get; private set; }
        internal static bool isDebugging { get; private set; }
        internal static string PrivateCryptoKey { get; private set; }
        internal static string PublicCryptoKey { get; private set; }
        internal static string StartupPath { get; private set; }
        internal static bool ShutdownInitiated { private set; get; }


        internal static void Initialize()
        {
			Console.WriteLine("Initializing environment...");
			Console.WriteLine();

            Stopwatch startWatch = Program.startWatch;

            StartupPath = AppDomain.CurrentDomain.BaseDirectory;
            isDebugging = Debugger.IsAttached;
            displayPacketActivity = isDebugging;
            DefaultEncoding = Encoding.Default;

            LanguageLocale.Init();

			try
			{
				PetCommandHandeler.Init();
				PetLocale.Init();
				LanguageLocale.InitSwearWord();
                LoadCommands();
				Stopwatch moduleWatch = new Stopwatch();

                if (!Directory.Exists(StartupPath + "/Logs"))
                    Directory.CreateDirectory(StartupPath + "/Logs");

				DirectoryInfo logDirInfo = new DirectoryInfo(StartupPath + "/Logs");
				foreach (FileInfo file in logDirInfo.GetFiles())
				{
					file.Delete();
				}

                LoadConfiguration();

				Game = new Game();
				Console.WriteLine("Pre-init -> ({0} ms)", startWatch.ElapsedMilliseconds);
			
				Console.Write("Connecting to database -> ");
				moduleWatch.Start();

                DatabaseManager = new DatabaseManager(Configuration.Database.Host, Configuration.Database.Port, 
                            Configuration.Database.Username, Configuration.Database.Password, 
                            Configuration.Database.DatabaseName, Configuration.Database.MinPoolSize, 
                            Configuration.Database.MaxPoolSize);
				DatabaseManager.init();

				using (IQueryAdapter dbClient = DatabaseManager.getQueryreactor())
				{
					Console.WriteLine("({0} ms)", moduleWatch.ElapsedMilliseconds);
					moduleWatch.Restart();
					Game.ContinueLoading(dbClient);
					moduleWatch.Restart();

					ConnectionManager = new ConnectionHandeling(Configuration.Networking.GamePort,
                            Configuration.Networking.GamePortConnectionLimit, Configuration.Networking.EnableAglesAlgorithm);
					ConnectionManager.Start();

                    if (Configuration.WatchdogEnabled)
                    {
                        DataSocket.SetupListener(Configuration.Networking.MonitoringPort);
					    DataSocket.Start();
                    }
					
					Console.WriteLine("Socket -> READY! ({0} ms)", moduleWatch.ElapsedMilliseconds);
					moduleWatch.Restart();

					StaticClientMessageHandler.Initialize();
					ClientMessageFactory.Init();

					mus = new MusSocket(Configuration.Networking.MusPort, Configuration.Networking.AllowedMusIPAddresses);
					mus.WaitForConnections();
					
					DataRow dRow;

					dbClient.setQuery("SELECT * FROM banner LIMIT 1");
					dRow = dbClient.getRow();


					PublicCryptoKey = (string)dRow[0];
					PrivateCryptoKey = (string)dRow[1];
					Console.WriteLine("MUS socket -> READY! ({0} ms)", moduleWatch.ElapsedMilliseconds);
				}

				Logging.WriteLine("Butterfly -> Ready to fly! (" + startWatch.ElapsedMilliseconds + " ms)");
                ServerStarted = DateTime.Now;

				if (ButterflyEnvironment.isDebugging)
				{
					Console.ForegroundColor = ConsoleColor.Yellow;
					Logging.WriteLine("Server is debugging: Console writing enabled");
					Console.ForegroundColor = ConsoleColor.White;
					Logging.DisabledState = false;
				}
				else
				{
					Logging.WriteLine("Server is not debugging: Console writing disabled");
					Logging.DisablePrimaryWriting(false);
				}

				Console.Title = VersionString;
			}
			catch (ModuleInitializationException e)
			{
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("Failed to initialize Butterfly Emulator:");
				Console.WriteLine(e.Message);
				Console.WriteLine();
				Console.WriteLine("Press any key to exit");
				Console.ReadKey();
				Environment.Exit(0);
			}
			catch (KeyNotFoundException e)
			{
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("Please check your configuration file - some values appear to be missing.");
				Console.WriteLine("Press any key to shut down ...");
				Console.WriteLine(e.ToString());
				Console.ReadKey(true);
				Environment.Exit(0);
			}
			catch (InvalidOperationException e)
			{
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("Failed to initialize ButterflyEmulator: " + e.Message);
				Console.WriteLine("Press any key to shut down ...");

				Console.ReadKey(true);
				Environment.Exit(0);
			}

			catch (Exception e)
			{
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("Fatal error during startup: " + e.ToString());
				Console.WriteLine("Press a key to exit");

				Console.ReadKey();
				Environment.Exit(1);
			}
        }
        
        internal static void LoadConfiguration()
        {
            Configuration = JsonReader<ButterflyConfiguration>.FromFile("Settings/Configuration.json");
        }

        internal static void PreformShutDown()
        {
            ButterflyEnvironment.PreformShutDown(true);
        }

        internal static void PreformShutDown(bool exitWhenDone)
        {
            if (ShutdownInitiated)
                return;

            ShutdownInitiated = true;

            StringBuilder builder = new StringBuilder();
            DateTime shutdownStart = DateTime.Now;

            Game.GetClientManager().SendMassMessage(LanguageLocale.GetValue("shutdown.alert"));
            TextHandling.AppendTimeStampWithComment(ref builder, shutdownStart, "Hotel pre-warning");

            Game.StopGameLoop();
            Console.Write("Game loop stopped");

            DateTime connectionClose = DateTime.Now;
            Console.WriteLine("Server shutting down...");
            Console.Title = "<<- SERVER SHUTDOWN ->>";
            
            ConnectionManager.Destroy();
            TextHandling.AppendTimeStampWithComment(ref builder, connectionClose, "Socket close");

            DateTime connectionFinalizer = DateTime.Now;
            Game.GetClientManager().CloseAll();
            TextHandling.AppendTimeStampWithComment(ref builder, connectionFinalizer, "Furni pre-save and connection close");

            DateTime roomDestroyer = DateTime.Now;
            Console.WriteLine("<<- SERVER SHUTDOWN ->> ROOM SAVE");
            Game.GetRoomManager().RemoveAllRooms();
            TextHandling.AppendTimeStampWithComment(ref builder, roomDestroyer, "Room destructor");

            DateTime dbSave = DateTime.Now;

            using (IQueryAdapter dbClient = DatabaseManager.getQueryreactor())
            {
               // dbClient.runFastQuery("TRUNCATE TABLE user_tickets");
                dbClient.runFastQuery("TRUNCATE TABLE user_online");
                dbClient.runFastQuery("TRUNCATE TABLE room_active");
                //dbClient.runFastQuery("UPDATE users SET online = 0");
                //dbClient.runFastQuery("UPDATE rooms SET users_now = 0");
            }
            TextHandling.AppendTimeStampWithComment(ref builder, dbSave, "Database pre-save");

            DateTime connectionShutdown = DateTime.Now;
            ConnectionManager.Destroy();
            TextHandling.AppendTimeStampWithComment(ref builder, connectionShutdown, "Connection shutdown");

            DateTime gameDestroy = DateTime.Now;
            Game.Destroy();
            TextHandling.AppendTimeStampWithComment(ref builder, gameDestroy, "Game destroy");

            DateTime databaseDeconstructor = DateTime.Now;

            try
            {
                Console.WriteLine("Destroying database manager.");
                DatabaseManager.destroy();
            }
            catch { }
            TextHandling.AppendTimeStampWithComment(ref builder, databaseDeconstructor, "Database shutdown");

            TimeSpan timeUsedOnShutdown = DateTime.Now - shutdownStart;
            builder.AppendLine("Total time on shutdown " + TextHandling.TimeSpanToString(timeUsedOnShutdown));
            builder.AppendLine("You have reached ==> [END OF SESSION]");
            builder.AppendLine();
            builder.AppendLine();
            builder.AppendLine();

            Logging.LogShutdown(builder);

            Console.WriteLine("System disposed, goodbye!");
            if (exitWhenDone)
                Environment.Exit(Environment.ExitCode);
        }

        internal static void LoadCommands()
        {
            CommandInvoker = new CommandInvoker();
        }
    }
}
