﻿using Butterfly.Core;
using Butterfly.HabboHotel.Achievements;
using Butterfly.HabboHotel.Catalogs;
using Butterfly.HabboHotel.GameClients;
using Butterfly.HabboHotel.Group;
using Butterfly.HabboHotel.Items;
using Butterfly.HabboHotel.Misc;
using Butterfly.HabboHotel.Navigators;
using Butterfly.HabboHotel.Quests;
using Butterfly.HabboHotel.Roles;
using Butterfly.HabboHotel.RoomBots;
using Butterfly.HabboHotel.Rooms;
using Butterfly.HabboHotel.SoundMachine;
using Butterfly.HabboHotel.Support;
using Butterfly.HabboHotel.Users.Inventory;
using Database_Manager.Database.Session_Details.Interfaces;
using System;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;


namespace Butterfly.HabboHotel
{
    class Game
    {
        #region Fields
        private GameClientManager ClientManager;
        private ModerationBanManager BanManager;
        private RoleManager RoleManager;
        private HelpTool HelpTool;
        private Catalog Catalog;
        private Navigator Navigator;
        private ItemManager ItemManager;
        private RoomManager RoomManager;
        private PixelManager PixelManager;
        private AchievementManager AchievementManager;
        private ModerationTool ModerationTool;
        private BotManager BotManager;
		private GroupManager GroupManager;
        private InventoryGlobal globalInventory;
        private QuestManager questManager;
        private HabboDataManager habboManager;

        private Task gameLoop;
        private bool gameLoopActive;
        private bool gameLoopEnded;
        private const int gameLoopSleepTime = 25;
		private Stopwatch moduleWatch;

        #endregion

        #region Return values
        internal HabboDataManager GetHabboDataManager()
        {
            return habboManager;
        }

        internal GameClientManager GetClientManager()
        {
            return ClientManager;
        }

        internal ModerationBanManager GetBanManager()
        {
            return BanManager;
        }

        internal RoleManager GetRoleManager()
        {
            return RoleManager;
        }

        internal HelpTool GetHelpTool()
        {
            return HelpTool;
        }

        internal Catalog GetCatalog()
        {
            return Catalog;
        }

        internal Navigator GetNavigator()
        {
            return Navigator;
        }

        internal ItemManager GetItemManager()
        {
            return ItemManager;
        }

        internal RoomManager GetRoomManager()
        {
            return RoomManager;
        }

		internal GroupManager GetGroupManager()
		{
			return GroupManager;
		}

        internal PixelManager GetPixelManager()
        {
            return PixelManager;
        }

        internal AchievementManager GetAchievementManager()
        {
            return AchievementManager;
        }

        internal ModerationTool GetModerationTool()
        {
            return ModerationTool;
        }

        internal BotManager GetBotManager()
        {
            return BotManager;
        }

        internal InventoryGlobal GetInventory()
        {
            return globalInventory;
        }

        internal QuestManager GetQuestManager()
        {
            return questManager;
        }
        #endregion

        #region Boot
        internal Game()
        {
            ClientManager = new GameClientManager();
            BanManager = new ModerationBanManager();
            habboManager = new HabboDataManager();
            RoleManager = new RoleManager();
            HelpTool = new HelpTool();
            Catalog = new Catalog();
            Navigator = new Navigator();
            ItemManager = new ItemManager();
            RoomManager = new RoomManager();
            PixelManager = new PixelManager();
			GroupManager = new Group.GroupManager();

            ModerationTool = new ModerationTool();
            BotManager = new BotManager();
            questManager = new QuestManager();
			moduleWatch = new Stopwatch();
        }

        internal void ContinueLoading(IQueryAdapter dbClient)
        {
            Stopwatch watch = new Stopwatch();
            watch.Start();

            ItemManager.LoadItems(dbClient);
            globalInventory = new InventoryGlobal();
            Logging.WriteLine("Item manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            Catalog.Initialize(dbClient);
            Logging.WriteLine("Catalog -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            Catalog.InitCache();
            Logging.WriteLine("Catalogue manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            LanguageLocale.LoadSwearWords(dbClient);
            BanManager.LoadBans(dbClient);
            Logging.WriteLine("Ban manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            RoleManager.LoadRights(dbClient);
            Logging.WriteLine("Role manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            HelpTool.LoadCategories(dbClient);
            HelpTool.LoadTopics(dbClient);
            Logging.WriteLine("Help tool -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            Navigator.Initialize(dbClient);
            Logging.WriteLine("Navigator -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            RoomManager.LoadModels(dbClient);
            RoomManager.InitRoomLinks(dbClient);
            RoomManager.InitVotedRooms(dbClient);
            Logging.WriteLine("Room manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            AchievementManager = new AchievementManager(dbClient);
            questManager.Initialize(dbClient);
            Logging.WriteLine("Achievement manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            ModerationTool.LoadMessagePresets(dbClient);
            ModerationTool.LoadPendingTickets(dbClient);
            Logging.WriteLine("Moderation tool -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            BotManager.LoadBots(dbClient);
            Logging.WriteLine("Bot manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            SongManager.Initialize();
            Logging.WriteLine("Song manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            GroupManager.InitGroups(dbClient);
            Logging.WriteLine("Group manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
            watch.Restart();

            DatabaseCleanup(dbClient);
            LowPriorityWorker.Init(dbClient);
            Logging.WriteLine("Database -> Cleanup performed! (" + watch.ElapsedMilliseconds + " ms)");
			watch.Restart();

            StartGameLoop();

			Logging.WriteLine("Game manager -> READY! (" + watch.ElapsedMilliseconds + " ms)");
        }

        #endregion

        #region Game loop
        internal void StartGameLoop()
        {
            gameLoopEnded = false;
            gameLoopActive = true;
            gameLoop = new Task(MainGameLoop);
            gameLoop.Start();
        }

        internal void StopGameLoop()
        {
            gameLoopActive = false;

            while (!gameLoopEnded)
            {
                Thread.Sleep(gameLoopSleepTime);
            }
        }

        internal byte GameLoopStatus = 0;

        internal static bool gameLoopEnabled = true;
        private void MainGameLoop()
        {
            while (gameLoopActive)
            {
                if (gameLoopEnabled)
                {
                    try
                    {
						moduleWatch.Restart();
                        LowPriorityWorker.Process(); //1 query
						if (moduleWatch.ElapsedMilliseconds > 10000)
							Console.WriteLine("High latency in LowPriorityWorker.Process ({0} ms)", moduleWatch.ElapsedMilliseconds);

						moduleWatch.Restart();

                        LowPriorityWorker.ConsoleTitleWorker();
						if (moduleWatch.ElapsedMilliseconds > 10000)
							Console.WriteLine("High latency in ConsoleTitleWorker.Process ({0} ms)", moduleWatch.ElapsedMilliseconds);

						moduleWatch.Restart();

						RoomManager.OnCycle(moduleWatch); // Queries for furni save
						ClientManager.OnCycle(moduleWatch);
                    }
                    catch (Exception e)
                    {
                        Logging.LogCriticalException("Exception in main environment thread: " + e.ToString());
                    }
                    GameLoopStatus = 8;
                }

                Thread.Sleep(gameLoopSleepTime);
            }

            gameLoopEnded = true;
        }
        #endregion

        #region Shutdown
        internal static void DatabaseCleanup(IQueryAdapter dbClient)
        {
            //dbClient.runFastQuery("TRUNCATE TABLE user_tickets");
            dbClient.runFastQuery("TRUNCATE TABLE user_online");
            dbClient.runFastQuery("TRUNCATE TABLE room_active");
            dbClient.runFastQuery("UPDATE server_status SET status = 1, users_online = 0, rooms_loaded = 0, server_ver = '" + ButterflyEnvironment.VersionString + "', stamp = '" + TextHandling.GetUnixTimestamp() + "' ");
        }

        internal void Destroy()
        {
            using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
            {
                DatabaseCleanup(dbClient);
            }

            Console.WriteLine("Destroyed Habbo Hotel.");
        }
        #endregion

        
    }
}
