﻿using Butterfly.Collections;
using Butterfly.Core;
using Butterfly.HabboHotel.Misc;
using Butterfly.HabboHotel.Users;
using Butterfly.HabboHotel.Users.Authenticator;
using Butterfly.HabboHotel.Users.Messenger;
using Butterfly.Messages;
using Butterfly.Messages.IOCodes;
using ConnectivityAsync;
using Database_Manager.Database.Session_Details.Interfaces;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Linq;

namespace Butterfly.HabboHotel.GameClients
{
    class GameClientManager
    {
        #region Fields
        private QueuedDictionary<uint, GameClient> clients;
        
        private Hashtable usernameRegister;
        private Hashtable userIDRegister;

        private Stopwatch checkCycleUpdatesWatch;
        private Stopwatch cycleCreditsLastUpdateWatch;
        private Stopwatch checkEffectStopwatch;
        private Stopwatch clientPingStopwatch;
        private PingHandler pinger;
        #endregion

        #region Return values
        internal GameClient GetClientByUserID(uint userID)
        {
            return userIDRegister[userID] as GameClient;
        }

        internal GameClient GetClientByUsername(string username)
        {
            return usernameRegister[username.ToLower()] as GameClient;
        }

        internal GameClient GetClient(uint clientID)
        {
            GameClient client = null;
            clients.Inner.TryGetValue(clientID, out client);
            return client;
        }

        internal bool ClientIsOnline(uint clientID)
        {
            return clients.ContainsKey(clientID);
        }

        internal int ClientCount
        {
            get
            {
                return clients.Inner.Count;
            }
        }

        internal void CloseConnection(uint connectionId)
        {
            this.clients.Remove(connectionId);
        }

        internal uint GetIdByName(string name)
        {
            GameClient client = GetClientByUsername(name);
            if (client != null)
            { 
                return client.GetHabbo().Id;
            }
            else
            { 
                using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
                {
                    dbClient.setQuery("SELECT id FROM users WHERE username = @username");
                    dbClient.addParameter("username", name);
                    return (uint)dbClient.getInteger();
                }
            }
        }

        internal string GetNameById(uint Id)
        {
            GameClient client = GetClientByUserID(Id);

            if (client != null)
                return client.GetHabbo().Username;

            string username;

            using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
            {
                dbClient.setQuery("SELECT username FROM users WHERE id = " + Id);
                username = dbClient.getString();
            }

            return username;
        }

        internal IEnumerable<GameClient> GetClientsById(Dictionary<uint, MessengerBuddy>.KeyCollection users)
        {
            foreach (uint id in users)
            {
                GameClient client = GetClientByUserID(id);
                if (client != null)
                    yield return client;
            }
        }

        #endregion

        #region Constructor
        internal GameClientManager()
        {
            this.pinger = new PingHandler();
            this.clients = new QueuedDictionary<uint, GameClient>(new EventHandler(AddClient), null, null, null);
			this.usernameRegister = new Hashtable(10103);
			this.userIDRegister = new Hashtable(10103);

			this.checkEffectStopwatch = new Stopwatch();
            this.checkCycleUpdatesWatch = new Stopwatch();
            this.cycleCreditsLastUpdateWatch = new Stopwatch();
            this.clientPingStopwatch = new Stopwatch();

            checkCycleUpdatesWatch.Start();
            cycleCreditsLastUpdateWatch.Start();
            clientPingStopwatch.Start();
			checkEffectStopwatch.Start();
        }
        #endregion

        #region Threads

        internal void OnCycle(Stopwatch watch)
        {
            CheckCronTasks(); // Pixel lolz and credits
			HandleFunctionReset(watch, "CheckCycleUpdates");
            TestClientConnections();
			HandleFunctionReset(watch, "TestClientConnections");
            clients.OnCycle();
        }

		private void HandleFunctionReset(Stopwatch watch, string methodName)
		{
			if (watch.ElapsedMilliseconds > 10000)
				Console.WriteLine("High latency in {0}: {1}ms", methodName, watch.ElapsedMilliseconds);
			watch.Restart();
		}

        private void CheckCronTasks()
        {
            ExecuteCron(checkCycleUpdatesWatch, ButterflyEnvironment.Configuration.PixelCronCycleTime, (GameClient client) => 
            {
                PixelManager.GivePixels(client, ButterflyEnvironment.Configuration.PixelCronAmount);
            });

            ExecuteCron(checkCycleUpdatesWatch, ButterflyEnvironment.Configuration.CreditsCronCycleTime, (GameClient client) => 
            {
                client.GetHabbo().Credits += ButterflyEnvironment.Configuration.CreditsCronAmount;
                client.GetHabboDataContainer().UpdateCreditsBalance();
            });

			if (checkEffectStopwatch.ElapsedMilliseconds > 3600)
			{
				checkEffectStopwatch.Restart();
				CheckEffects();
			}
        }

        private void ExecuteCron(Stopwatch watch, uint cycleTime, Action<GameClient> cronAction)
        {
            try
            { 
                if (watch.ElapsedMilliseconds > cycleTime * 1000)
                {
                    watch.Restart();

                    foreach (GameClient client in clients.Values)
                    {
                        if (client.GetHabbo() != null)
                            cronAction(client);
                    }
                }
            }
            catch (Exception ex)
            {
                Logging.LogException("Error during cron purge: " + ex.ToString());
            }
        }

        private void TestClientConnections()
        {
            if (pinger.IsTimeForPing())
            {
                try
                {
                    foreach (GameClient client in clients.Values)
                    {
                        pinger.CheckClientState(client);
                    }
                    
                    pinger.SendPings();
                }
                catch (Exception e) { Logging.LogThreadException(e.ToString(), "Ping"); }
            }
        }
        #endregion

        #region Collection modyfying
        
        private void AddClient(object obj, EventArgs args)
        {
            KeyValuePair<uint, GameClient> addedPair = (KeyValuePair<uint, GameClient>)obj;
            addedPair.Value.StartConnection();
        }
        
        #endregion

        #region Methods
        

        internal void CreateAndStartClient(uint clientID, ConnectionInformation connection)
        {
            GameClient client = new GameClient(clientID, connection);
            clients.Add(clientID, client);
        }

        internal void QueueBroadcaseMessage(ServerMessage message)
        {
            clients.RequestAction((GameClient client) =>
            {
                client.SendMessageWithoutReturning(message);
            }, message.ReturnToOwner);
        }

        internal void QueueBroadcaseMessage(ServerMessage message, string requirements)
        {
            clients.RequestAction((GameClient client) => 
            {
                if (client.GetHabboDataContainer() != null && client.GetHabboDataContainer().HasFuse(requirements))
                    client.SendMessageWithoutReturning(message);
            }, message.ReturnToOwner);
        }

        private void BroadcastMessage(ServerMessage message)
        {
            clients.RequestAction((GameClient client) =>
            {
                client.SendMessageWithoutReturning(message);
            }, message.ReturnToOwner);
        }

        internal void QueueCreditsUpdate(int amount)
        {
            clients.RequestAction((GameClient client) =>
            {
                if (client.GetHabbo() != null)
                {
                    client.GetHabbo().Credits += amount;
                    client.GetHabboDataContainer().UpdateCreditsBalance();
                    client.SendNotif(LanguageLocale.GetValue("user.creditsreceived") + ": " + amount);
                }
            });
        }

        internal void QueueBadgeUpdate(string badge)
        {
            clients.RequestAction((GameClient client) =>
            {
                if (client.GetHabbo() != null)
                {
                    client.GetHabboDataContainer().GetBadgeComponent().GiveBadge(badge, true);
                    client.SendNotif(LanguageLocale.GetValue("user.badgereceived"));
                }
            });
        }

        internal void QueuePixelUpdate(int amount)
        {
            clients.RequestAction((GameClient client) =>
            {
                PixelManager.GivePixels(client, amount);
            });
        }

        internal void QueueStaffMessage(string message, int minRank = 3)
        {
            clients.RequestAction((GameClient client) =>
            {
                if (client.GetHabbo() != null && client.GetHabbo().Rank > minRank)
                    client.SendNotif("Staff message: " + message);
            });
        }
        
        private void CheckEffects()
        {
            foreach (GameClient client in clients.Values)
            {
                if (client.GetHabbo() == null || client.GetHabboDataContainer().GetAvatarEffectsInventoryComponent() == null)
                    continue;

                client.GetHabboDataContainer().GetAvatarEffectsInventoryComponent().CheckExpired();
            }
        }

        internal void LogClonesOut(uint UserID)
        {
            GameClient client = GetClientByUserID(UserID);
            if (client != null)
                client.Disconnect();
        }

        internal void RegisterClient(GameClient client, uint userID, string username)
        {
            if (usernameRegister.ContainsKey(username.ToLower()))
                usernameRegister[username.ToLower()] = client;
            else
                usernameRegister.Add(username.ToLower(), client);

            if (userIDRegister.ContainsKey(userID))
                userIDRegister[userID] = client;
            else
                userIDRegister.Add(userID, client);

            using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
            {
                dbClient.setQuery("REPLACE INTO user_online VALUES (" + userID + ")");
            }
        }

        internal void UnregisterClient(uint userid, string username)
        {
            userIDRegister.Remove(userid);
            usernameRegister.Remove(username.ToLower());

            using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
            {
                dbClient.setQuery("DELETE FROM user_online WHERE id = " + userid);
            }
        }

        internal void CloseAll()
        {
            StringBuilder queryBuilder = new StringBuilder();
            int clientSaveCount = clients.Inner.Count(t => t.Value.GetHabbo() != null);
            int current = 0;
            int clientLength = ClientCount;

            foreach (GameClient client in clients.Values)
            {
                if (client.GetHabbo() != null)
                {
                    try
                    {
                        client.GetHabboDataContainer().GetInventoryComponent().RunDBUpdate();
                        queryBuilder.Append(client.GetHabbo().GetQueryString);

                        Console.Clear();
                        Console.WriteLine("<<- SERVER SHUTDOWN ->> IVNENTORY SAVE: " + String.Format("{0:0.##}", ((double)++current / clientSaveCount) * 100) + "%");
                    }
                    catch { }
                }
            }

            if (queryBuilder.Length > 0)
            {
                using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
                {
                    dbClient.runFastQuery(queryBuilder.ToString());
                }
            }

            Console.WriteLine("Done saving users inventory!");
            Console.WriteLine("Closing server connections...");
            try
            {
                int i = 0;
                foreach (GameClient client in clients.Values)
                {
                    i++;
                    if (client.GetConnection() != null)
                    {
                        try
                        {
                            client.GetConnection().Destroy();
                        }
                        catch { }

                        Console.Clear();
                        Console.WriteLine("<<- SERVER SHUTDOWN ->> CONNECTION CLOSE: " + String.Format("{0:0.##}", ((double)i / clientLength) * 100) + "%");
                    }
                }
            }
            catch (Exception e)
            {
                Logging.LogCriticalException(e.ToString());
            }
            clients.Clear();
            pinger.Destroy();
            Console.WriteLine("Connections closed!");
        }

        internal void SendMassMessage(string text)
        {
            try
            {
                ServerMessage message = PacketHandelingManager.GetRecycledItem(OutputCode.HabboBroadcastMessageEvent);
                message.AddString(text);
                QueueBroadcaseMessage(message);
            }
            catch (Exception e) { Logging.HandleException(e, "GameClientManager.SendMassMessage"); }
        }
        #endregion
    }
}
