﻿using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using Butterfly.HabboHotel.Items;
using Butterfly.HabboHotel.Rooms.Games;
using Butterfly.HabboHotel.Rooms.Wired.WiredHandlers.Interfaces;
using Butterfly.Messages;
using WiredSolver;
using System.Threading;
using System;
using Butterfly.Messages.IOCodes;

namespace Butterfly.HabboHotel.Rooms.Wired
{
    public class WiredHandler
    {
        #region Declares
        private Hashtable actionStacks; //int, List<RoomItem>
        private Dictionary<RoomItem, int> oldLocations = new Dictionary<RoomItem, int>();
        private Queue requestedTriggers;
        private Queue requestingUpdates;

        private WiredSolverInstance wireSlower;

        private Room room;
        internal ConditionHandler conditionHandler;
        private bool doCleanup = false;
        #endregion

        #region Constructor
        public WiredHandler(Room room)
        {
            this.actionStacks = new Hashtable();
            this.requestedTriggers = new Queue();
            this.requestingUpdates = new Queue();
            this.wireSlower = new WiredSolverInstance();

            this.room = room;
            this.conditionHandler = new ConditionHandler(room);
        }
        #endregion

        #region Furniture add/remove
        internal void AddFurniture(RoomItem item)
        {
            AddFurnitureToItemStack(item);
        }

        internal void RemoveFurniture(RoomItem item)
        {
            RemoveFurnitureFromStack(item);
        }

        private void AddFurnitureToItemStack(RoomItem item)
        {
            int itemCoord = item.GetDoubleCoordinate();

            int cord;
            if (oldLocations.TryGetValue(item, out cord))
            {
                if (cord == itemCoord)
                    return;
                ((List<RoomItem>)actionStacks[cord]).Remove(item);
                oldLocations.Remove(item);
            }
			
            if (actionStacks.ContainsKey(itemCoord))
            {
                oldLocations.Add(item, itemCoord);
                ((List<RoomItem>)actionStacks[itemCoord]).Add(item);
            }
            else
            {
                List<RoomItem> stack = new List<RoomItem>();
                stack.Add(item);
                actionStacks.Add(itemCoord, stack);
                oldLocations.Add(item, itemCoord);
            }
        }

        private void RemoveFurnitureFromStack(RoomItem item)
        {
            int cord;
            if (oldLocations.TryGetValue(item, out cord))
            {
                ((List<RoomItem>)actionStacks[cord]).Remove(item);
                oldLocations.Remove(item);
            }
        }
        #endregion

        #region Room cycle management
        internal void OnCycle()
        {
            if (doCleanup)
            {
                foreach (List<RoomItem> list in actionStacks.Values)
                {
                    foreach (RoomItem item in list)
                    {
                        if (item.wiredCondition != null)
                        {
                            item.wiredCondition.Dispose();
                            item.wiredCondition = null;
                        }
                        if (item.wiredHandler != null)
                        {
                            item.wiredHandler.Dispose();
                            item.wiredHandler = null;
                        }
                    }
                }
                actionStacks.Clear();
                requestedTriggers.Clear();
                requestingUpdates.Clear();

                doCleanup = false;
                return;
            }
            if (requestingUpdates.Count > 0)
            {
                List<IWiredCycleable> toAdd = new List<IWiredCycleable>();
                lock (requestingUpdates.SyncRoot)
                {
                    while (requestingUpdates.Count > 0)
                    {
                        IWiredCycleable handler = (IWiredCycleable)requestingUpdates.Dequeue();
                        if (handler.Disposed())
                        {
                            continue;
                        }

                        if (handler.OnCycle())
                        {
                            toAdd.Add(handler);
                        }
                    }

                    foreach (IWiredCycleable cycle in toAdd)
                    {
                        requestingUpdates.Enqueue(cycle);
                    }
                }
            }

            conditionHandler.OnCycle();
        }
        #endregion

        #region Functions
        internal void OnPickall()
        {
            doCleanup = true;
        }

        internal void OnEvent(uint itemID)
        {
            //ServerMessage message = new ServerMessage(OutputCode.ObjectDataUpdateMessageEvent);
            //message.AddString(itemID.ToString());
            //message.AddString("0");

            //room.SendMessage(message);
        }

        internal void Reset()
        {
            this.actionStacks.Clear();
            requestedTriggers.Clear();
            requestingUpdates.Clear();
            wireSlower = new WiredSolverInstance();
        }
        #endregion

        #region Requests
       
        internal void RequestStackHandle(int x, int y, RoomItem item, RoomUser user, Team team, Room theRoom)
        {   
			int coordinate = TextHandling.CombineXYCoord(x, y);
            IEnumerable<RoomItem> items = GetWiredStack(coordinate); 
            if (items != null && conditionHandler.AllowsHandling(x,y, user, theRoom))
            {
                foreach (RoomItem stackItem in items)
                {
                    if (stackItem.wiredHandler is IWiredEffect)
                    {
                        ((IWiredEffect)stackItem.wiredHandler).Handle(user, team, item);
                    }
                }

                bool shouldBeHandled = false;
				coordinate = TextHandling.CombineXYCoord(x, y + 1);

				CheckHandlingState(ref shouldBeHandled, ref coordinate);

				if (!shouldBeHandled)
				{
					coordinate = TextHandling.CombineXYCoord(x + 1, y);
					CheckHandlingState(ref shouldBeHandled, ref coordinate);
				}

				if (!shouldBeHandled)
				{
					coordinate = TextHandling.CombineXYCoord(x, y - 1);
					CheckHandlingState(ref shouldBeHandled, ref coordinate);
				}

				if (!shouldBeHandled)
				{
					coordinate = TextHandling.CombineXYCoord(x - 1, y);
					CheckHandlingState(ref shouldBeHandled, ref coordinate);
				}

                if (shouldBeHandled)
                {
                    room.GetWiredHandler().TriggerOnWire(x, y);
                }
            }
        }

        private IEnumerable<RoomItem> GetWiredStack(int coordinate)
        {
            List<RoomItem> result = actionStacks[coordinate] as List<RoomItem>;
            if (result != null)
            {
                return result.ToArray();
            }
            return null;
        }
        private void CheckHandlingState(ref bool shouldBeHandled, ref int coordinate)
        {
            foreach (RoomItem _item in room.GetGameMap().GetCoordinatedItems(ref coordinate))
            {
                if (WiredHandler.TypeIsWire(_item.GetBaseItem().InteractionType))
                {
                    shouldBeHandled = true;
                    return;
                }
            }
        }

        internal void RequestCycle(IWiredCycleable handler)
        {
            lock (requestingUpdates.SyncRoot)
            {
                requestingUpdates.Enqueue(handler);
            }
        }

        #endregion

        #region Return values
        internal Room GetRoom()
        {
            return room;
        }

        #endregion

        #region Wires

        internal void AddWire(RoomItem item, Point location, int rotation, InteractionType wireType)
        {

            WireCurrentTransfer transfer = WireCurrentTransfer.NONE;

            switch (wireType)
            {
                case InteractionType.wireCenter:
                    {
                        transfer = WireCurrentTransfer.DOWN | WireCurrentTransfer.LEFT | WireCurrentTransfer.RIGHT | WireCurrentTransfer.UP;
                        break;
                    }
                case InteractionType.wireCorner:
                    {
                        switch (rotation)
                        {
                            default:
                                {
                                    transfer = WireCurrentTransfer.DOWN | WireCurrentTransfer.RIGHT;
                                    break;
                                }
                            case 2:
                                {
                                    transfer = WireCurrentTransfer.DOWN | WireCurrentTransfer.LEFT;
                                    break;
                                }
                            case 4:
                                {
                                    transfer = WireCurrentTransfer.UP | WireCurrentTransfer.LEFT;
                                    break;
                                }
                            case 6:
                                {
                                    transfer = WireCurrentTransfer.RIGHT | WireCurrentTransfer.UP;
                                    break;
                                }
                        }

                        break;
                    }
                case InteractionType.wireSplitter:
                    {
                        switch (rotation)
                        {
                            default:
                                {
                                    transfer = WireCurrentTransfer.UP | WireCurrentTransfer.DOWN | WireCurrentTransfer.RIGHT;
                                    break;
                                }
                            case 2:
                                {
                                    transfer = WireCurrentTransfer.LEFT | WireCurrentTransfer.RIGHT | WireCurrentTransfer.DOWN;
                                    break;
                                }
                            case 4:
                                {
                                    transfer = WireCurrentTransfer.UP | WireCurrentTransfer.DOWN | WireCurrentTransfer.LEFT;
                                    break;
                                }
                            case 6:
                                {
                                    transfer = WireCurrentTransfer.LEFT | WireCurrentTransfer.RIGHT | WireCurrentTransfer.UP;
                                    break;
                                }
                        }
                        break;
                    }

                case InteractionType.wireStandard:
                    {
                        switch (rotation)
                        {
                            default:
                            case 0:
                            case 4:
                                {
                                    transfer = WireCurrentTransfer.UP | WireCurrentTransfer.DOWN;
                                    break;
                                }
                            case 2:
                            case 6:
                                {
                                    transfer = WireCurrentTransfer.LEFT | WireCurrentTransfer.RIGHT;
                                    break;
                                }
                        }
                        break;
                    }
            }

            if (transfer == WireCurrentTransfer.NONE)
                return;

			List<Point> pointsToTrigger = MarshallPoint(wireSlower.AddOrUpdateWire(location.X, location.Y, CurrentType.OFF, transfer));
            //form.AddOrUpdateWire(location.X, location.Y, CurrentType.OFF, transfer);
            HandleItems(pointsToTrigger);
        }

        internal void RemoveWiredItem(Point coordinate)
        {
            //TriggerOnWire(coordinate);
            wireSlower.RemoveWire(coordinate.X, coordinate.Y);
            //form.Remove(coordinate.X, coordinate.Y);
        }

        internal static bool TypeIsWire(InteractionType type)
        {
            switch (type)
            {
                case InteractionType.wireCenter:
                case InteractionType.wireCorner:
                case InteractionType.wireSplitter:
                case InteractionType.wireStandard:
                    {
                        return true;
                    }
                default:
                    return false;
            }
        }

        internal void TriggerOnWire(int x, int y)
        {
            WireTransfer transferItem = wireSlower.getWireTransfer(new System.Drawing.Point(x, y));
            
            List<Point> pointsToTrigger;
            WireCurrentTransfer transferType = WireCurrentTransfer.DOWN | WireCurrentTransfer.LEFT | WireCurrentTransfer.RIGHT | WireCurrentTransfer.UP;
            if (transferItem.isPowered() && transferItem.getTransfer() == transferType)
            {
				pointsToTrigger = MarshallPoint(wireSlower.RemoveWire(x, y));
                //form.Remove(location.X, location.Y);
            }
            else
            {
				pointsToTrigger = MarshallPoint(wireSlower.AddOrUpdateWire(x, y, CurrentType.SENDER, WireCurrentTransfer.NONE));

				
                //pointsToTrigger = 
                //form.AddOrUpdateWire(location.X, location.Y, CurrentType.SENDER, WireCurrentTransfer.NONE);
            }
            
            HandleItems(pointsToTrigger, x, y);
        }

		private List<Point> MarshallPoint(List<System.Drawing.Point> points)
		{
			List<Point> newlist = new List<Point>(points.Count);
			foreach (System.Drawing.Point p in points)
				newlist.Add(new Point(p.X, p.Y));
			return newlist;
		}

        private void HandleItems(List<Point> coords, int startX, int startY)
        {
            foreach (Point point in coords)
            {
                if (point.X != startX && point.Y != startY)
                    HandleItems(point);
            }
        }

        private void HandleItems(List<Point> coords)
        {
            foreach (Point point in coords)
            {
                HandleItems(point);
            }
        }

        private void HandleItems(Point coord)
        {
            WireTransfer transfer = wireSlower.getWireTransfer(new System.Drawing.Point(coord.X, coord.Y));
            List<RoomItem> items = new List<RoomItem>(room.GetGameMap().GetCoordinatedItems(coord));
            foreach (RoomItem item in items)
            {
                item.Interactor.OnTrigger(null, item, 0, true);
            }
            items.Clear();
            items = null;
        }

        #endregion

        #region Loading
        internal void LoadWired()
        {

        }
        #endregion

        #region Unloading
        internal void Destroy()
        {
            if (actionStacks != null)
                actionStacks.Clear();
            requestedTriggers.Clear();
            requestingUpdates.Clear();
            if (wireSlower != null)
                wireSlower.Destroy();
            wireSlower = null;
            room = null;
        }
        #endregion

        internal void OnItemRemove(RoomItem Item)
        {
            
        }
    }
}
