﻿using Collections.ModernCollections;

namespace Butterfly.HabboHotel.Pathfinding.MapManagement
{
    class Stackmap
    {
        private OrderedArray<double, StackItem>[,] map;
        private object syncRoot;
        private const double TOLLERANCE = 2.5;

        public Stackmap(int dimensionX, int dimensionY)
        {
            this.map = new OrderedArray<double, StackItem>[dimensionX, dimensionY];
            this.syncRoot = new object();
        }

        internal void AddItem(uint itemID, int x, int y, double z, double height, byte state)
        {
            lock (syncRoot)
            {
                if (map[x, y] == null)
                    map[x, y] = new OrderedArray<double, StackItem>(1, SortType.Ascending);

                map[x, y].Insert(z, new StackItem(height + z, z, itemID, state));
            }
        }

        internal void RemoveItem(uint itemID, int itemX, int itemY)
        {
            lock (syncRoot)
            {
                bool found = false;
                for (int i = 0; i < map[itemX, itemY].Count && !found; i++)
                {
                    StackItem entry = map[itemX, itemY].StorageArray[i].Value;
                    if (entry.ItemID == itemID)
                    {
                        map[itemX, itemY].RemoveItem(entry);
                        found = true;
                    }
                }
            }
        }

        internal bool OpenWindow(int x, int y, double z, double height)
        {
            lock (syncRoot)
            {
                for (int i = 0; i < map[x, y].Count; i++)
                {
                    StackItem entry = (StackItem)map[x, y].StorageArray[i].Value;
                    if ((z > entry.ZMin && z < entry.ZMax) || (z + height > entry.ZMin && z + height < entry.ZMax))
                        return false;
                }
            }

            return true;
        }
        
        internal double GetNewHeight(double currentHeight, double lowest, int x, int y)
        {
            double closestHeight = lowest;
            StackItem lastStackItem = new StackItem(lowest, lowest, 0, 1);
            double lastWindowHeight;
            bool found = false;

            lock (syncRoot)
            {
                if (map[x, y] != null)  // If there are items on this square
                {
                    for (int i = 0; i < map[x, y].Count; i++)
                    {
                        StackItem entry = (StackItem)map[x, y].StorageArray[i].Value;

                        if (entry.ZMin >= lastStackItem.ZMax && entry.ZMin >= lowest) // Means we are ending a new window
                        {
                            lastWindowHeight = entry.ZMin - lastStackItem.ZMax;

                            //if (closestHeight == 0 && lastWindowHeight >= TOLLERANCE) // First one doesn't fit
                            //{
                            //    found = true;
                            //}
                            
                            if (lastWindowHeight >= TOLLERANCE 
                                && ((GetDistance(lastWindowHeight, currentHeight) <= GetDistance(closestHeight, currentHeight) || !found))
                                && lastStackItem.SquareState != 0)
                            {
                                closestHeight = lastStackItem.ZMax; 
                                found = true;
                            }
                            
                            lastStackItem = entry;
                        }
                    }

                    if (!found || GetDistance(lastStackItem.ZMax, currentHeight) <= GetDistance(closestHeight, currentHeight)) // Check if top of stack is closest
                    {
                        if (lastStackItem.SquareState == 3)      // If it is a chair
                            closestHeight = lastStackItem.ZMin;
                        else
                            closestHeight = lastStackItem.ZMax;

                    }
                }
                else return lowest;
            }

            return closestHeight;
        }

        internal void TryGetSquareStatus(int x, int y, double z, ref byte state)
        {
            if (x < 0 || y < 0 || x > map.GetLength(0) - 1 || y > map.GetLength(1) - 1)
            {
                state = 0;
                return;
            }

            lock (syncRoot)
            {
                try
                {
                    if (map[x, y] != null && map[x, y].Count > 0)
                    {
                        StackItem closest = map[x, y].StorageArray[0].Value;

                        for (int i = 0; i < map[x, y].Count; i++)
                        {
                            StackItem entry = map[x, y].StorageArray[i].Value;
                            if ((entry.ZMax <= z && z - entry.ZMax <= z - closest.ZMax) || (entry.ZMin >= closest.ZMin && entry.ZMax > closest.ZMax))
                            {
                                if (entry.ZMin == closest.ZMin)
                                {
                                    if (entry.SquareState < closest.SquareState)
                                        closest = entry;

                                    if (closest.SquareState == 0 && entry.SquareState > 0)
                                        closest = entry;

                                    if (closest.SquareState == 1 && entry.SquareState == 3)
                                        closest = entry;

                                }
                                else
                                {
                                    closest = entry;
                                }
                            }
                        }

                        if (closest.ZMax <= z || closest.ZMin == z)
                            state = closest.SquareState;
                        else
                            state = 1;
                    }
                }
                catch
                {
                    return;
                }
            }
        }

        internal void Clear(int x, int y)
        {
            lock (syncRoot)
            {
                if (map[x, y] != null)
                    map[x, y].Clear();
            }
        }

        private static double GetDistance(double x, double y)
        {
            if (x > y)
                return x - y;
            else
                return y - x;
        }
    }
}
