﻿using Butterfly.Core;
using Butterfly.HabboHotel.Catalogs;
using Butterfly.HabboHotel.ChatMessageStorage;
using Butterfly.HabboHotel.GameClients;
using Butterfly.HabboHotel.Group;
using Butterfly.HabboHotel.Users.Badges;
using Butterfly.HabboHotel.Users.Group;
using Butterfly.HabboHotel.Users.Inventory;
using Butterfly.HabboHotel.Users.Messenger;
using Butterfly.HabboHotel.Users.Subscriptions;
using Butterfly.HabboHotel.Users.UserDataManagement;
using Butterfly.Messages;
using Butterfly.Messages.IOCodes;
using Database_Manager.Database.Session_Details.Interfaces;
using System;

namespace Butterfly.HabboHotel.Users
{
    class HabboDataContainer : IUpdateListener<Habbo>
    {
        private uint EquippedGroup { get { return this.data.EquippedGroup; } set { this.data.EquippedGroup = value; } }
        private bool IsGroupMember { get { return this.data.IsGroupMember; } }
        internal Habbo data { get; private set; }

        private SubscriptionManager SubscriptionManager;
        private HotelChatHandler Messenger;
        private HabboMessenger ChatMessenger;
        private BadgeComponent BadgeComponent;
        private InventoryComponent InventoryComponent;
        private AvatarEffectsInventoryComponent AvatarEffectsInventoryComponent;
        private ChatMessageManager chatMessageManager;
        private UserGroupManager groupManager;
        private GameClient mClient;
        private bool messengerInitialized;

        public HabboDataContainer(Habbo data)
        {
            this.data = data;
            this.data.RegisterUpdater(this);
        }

        public void Destroy()
        {
            if (this.data != null)
            {
                this.data.OnDisconnect(this.GetClient());
                this.data.RemoveUpdater(this);
                if (this.groupManager != null)
                {
                    this.groupManager.Clean();
                    this.groupManager = null;
                }
                this.data = null;

                if (ChatMessenger != null)
                    ChatMessenger.LogChatHandlerOut(Messenger);

                if (SubscriptionManager != null)
                    SubscriptionManager.Destroy();

                if (AvatarEffectsInventoryComponent != null)
                    AvatarEffectsInventoryComponent.Dispose();

                if (InventoryComponent != null)
                {
                    InventoryComponent.SetIdleState();
                    InventoryComponent.RunDBUpdate();
                }

                this.mClient = null;
            }
        }

        internal void Init(GameClient client, UserData data)
        {
            this.mClient = client;

            this.SubscriptionManager = new SubscriptionManager(this.data.Id, data);
            this.BadgeComponent = new BadgeComponent(this.data.Id, data);
            this.InventoryComponent = InventoryGlobal.GetInventory(this.data.Id, client, data);
            this.InventoryComponent.SetActiveState(client);
            this.AvatarEffectsInventoryComponent = new AvatarEffectsInventoryComponent(this.data.Id, client, data);
            this.data.quests = data.quests;
            this.chatMessageManager = new ChatMessageManager();

            this.ChatMessenger = MessengerManager.instance.GetUserMessenger(this.data.Id, false, data);
            this.ChatMessenger.Init(data.friends, data.requests);

            this.Messenger = new HotelChatHandler(mClient);
            this.ChatMessenger.AddOnlineState(this.Messenger);

            this.data.SpectatorMode = false;
            this.data.Disconnected = false;
            this.data.UsersRooms = data.rooms;

        }

        internal void LoadData(UserData data)
        {
            this.data.LoadAchievements(data.achievements);
            this.data.LoadFavorites(data.favouritedRooms);
            this.data.LoadTags(data.tags);
        }

        internal GroupMember GetEquippedGroup()
        {
            return data.GetEquippedGroup();
        }

        internal ChatMessageManager GetChatMessageManager()
        {
            return chatMessageManager;
        }

        internal void DestroyGroupMembership(uint groupID)
        {
            GroupMember m = this.groupManager.GetGroupMember(groupID);
            m.LeaveGroup();
        }

        internal UserGroupManager GetGroupManager()
        {
            return this.groupManager;
        }

        internal SubscriptionManager GetSubscriptionManager()
        {
            return SubscriptionManager;
        }

        internal HotelChatHandler GetMessenger()
        {
            return Messenger;
        }
        internal HabboMessenger GetChatMessenger()
        {
            return ChatMessenger;
        }

        internal BadgeComponent GetBadgeComponent()
        {
            return BadgeComponent;
        }

        internal InventoryComponent GetInventoryComponent()
        {
            return InventoryComponent;
        }

        internal AvatarEffectsInventoryComponent GetAvatarEffectsInventoryComponent()
        {
            return AvatarEffectsInventoryComponent;
        }

        internal override void ItemUpdated(Habbo updatable)
        {
        }

        internal void InitGroups()
        {
            this.groupManager = new UserGroupManager(data, this.mClient);
        }

        internal void InitMessenger()
        {
            if (messengerInitialized)
                return;

            messengerInitialized = true;
            GetChatMessenger().OnStatusChanged(false);
        }

        internal void UpdateCreditsBalance()
        {
            ServerMessage message = PacketHandelingManager.GetRecycledItem(OutputCode.CreditBalanceEvent);
            message.AddString(data.Credits.ToString() + ".0");
            mClient.SendMessage(message);
        }

        internal void UpdateActivityPointsBalance()
        {
            if (mClient == null)
                return;

            ServerMessage message = PacketHandelingManager.GetRecycledItem(OutputCode.HabboActivityPointNotificationMessageEvent);
            message.AddInt32(2);
            message.AddInt32(0);
            message.AddInt32(this.data.ActivityPoints);
            message.AddInt32(CatalogItem.CURRENCY_USED);
			message.AddInt32(this.data.Crystals);
			message.AddInt32(GetClient().GetHabbo().AchievementPoints);

			mClient.SendMessage(message);
        }

        internal void Mute()
        {
            if (!this.data.Muted)
            {
                GetClient().SendNotif(LanguageLocale.GetValue("moderation.muted"));
                this.data.Muted = true;
            }
        }

        internal void Unmute()
        {
            if (this.data.Muted)
            {
                GetClient().SendNotif(LanguageLocale.GetValue("moderation.unmuted"));
                this.data.Muted = false;
            }
        }

        internal void SerializeClub()
        {
            GameClient Client = GetClient();
            ServerMessage club = PacketHandelingManager.GetRecycledItem(OutputCode.SerializeClub);


            if (GetSubscriptionManager().HasSubscription(SubscriptionManager.SUBSCRIPTION_TYPE_VIP))
            {
                club.AddString(SubscriptionManager.SUBSCRIPTION_TYPE_VIP);
                Double Expire = GetSubscriptionManager().GetSubscription().ExpireTime;
                Double TimeLeft = Expire - TextHandling.GetUnixTimestamp();
                int TotalDaysLeft = (int)Math.Ceiling(TimeLeft / 86400);
                int MonthsLeft = TotalDaysLeft / 31;

                if (MonthsLeft >= 1) MonthsLeft--;

                club.AddInt32(TotalDaysLeft);// - (MonthsLeft * 31));
                club.AddInt32(2); // ??
                club.AddInt32(MonthsLeft);
                club.AddInt32(1); // type
                club.AddBoolean(true);
                club.AddBoolean(true);
                club.AddInt32(0); //Time left HC
                club.AddInt32(0); // days i have on hc
                club.AddInt32(500); // days i have on vip
            }
            else if (GetSubscriptionManager().HasSubscription(SubscriptionManager.SUBSCRIPTION_TYPE_CLUB))
            {
                club.AddString(SubscriptionManager.SUBSCRIPTION_TYPE_CLUB);
                Double Expire = GetSubscriptionManager().GetSubscription().ExpireTime;
                Double TimeLeft = Expire - TextHandling.GetUnixTimestamp();
                int TotalDaysLeft = (int)Math.Ceiling(TimeLeft / 86400);
                int MonthsLeft = TotalDaysLeft / 31;

                if (MonthsLeft >= 1) MonthsLeft--;

                club.AddInt32(0);// - (MonthsLeft * 31));
                club.AddInt32(0); // ??
                club.AddInt32(0);
                club.AddInt32(1); // type
                club.AddBoolean(false);
                club.AddBoolean(true);
                club.AddInt32(0); //Time left HC
                club.AddInt32(0); // days i have on hc
                club.AddInt32(0); // days i have on vip
            }
            else
            {
                club.AddString(SubscriptionManager.SUBSCRIPTION_TYPE_CLUB);
                club.AddInt32(0);
                club.AddInt32(0); // ??
                club.AddInt32(0);
                club.AddInt32(1); // type
                club.AddBoolean(false);
                club.AddBoolean(true);
                club.AddInt32(0);
                club.AddInt32(0); // days i have on hc
                club.AddInt32(0); // days i have on vip
            }

            Client.SendMessage(club);

            ServerMessage fuserights = PacketHandelingManager.GetRecycledItem(OutputCode.FuserightEvent);
            if (GetSubscriptionManager().HasSubscription(SubscriptionManager.SUBSCRIPTION_TYPE_CLUB)) // VIP 
                fuserights.AddInt32(2);
            else
                fuserights.AddInt32(0);
            fuserights.AddInt32((int)this.data.Rank);

            Client.SendMessage(fuserights);
        }

        private GameClient GetClient()
        {
            return this.mClient;
        }

        internal bool HasFuse(string Fuse)
        {
            if (ButterflyEnvironment.Game.GetRoleManager().RankHasRight(data.Rank, Fuse))
            {
                return true;
            }

            if (GetSubscriptionManager().HasSubscription() && ButterflyEnvironment.Game.GetRoleManager().SubHasRight(GetSubscriptionManager().GetSubscription().SubscriptionId, Fuse))
            {
                return true;
            }

            return false;
        }
		
		internal void SynchronizeCrystalBalance()
		{
			int newBalance;
			using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
			{
				dbClient.setQuery("SELECT crystals FROM users WHERE id = @userid");
				dbClient.addParameter("userid", data.Id);
				newBalance = dbClient.getInteger();
			}

			if (newBalance != data.Crystals)
			{
				data.Crystals = newBalance;
				UpdateActivityPointsBalance();
			}
		}

        internal void StoreCrystalsNow()
        {
            using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
            {
                dbClient.setQuery("UPDATE users SET crystals = " + data.Crystals + " WHERE id = @userid");
                dbClient.addParameter("userid", data.Id);
                dbClient.runQuery();
            }
        }

        internal String getIP()
        {
            using (IQueryAdapter dbClient = ButterflyEnvironment.DatabaseManager.getQueryreactor())
            {
                dbClient.setQuery("SELECT ip_last FROM users WHERE  id = @userid");
                dbClient.addParameter("userid", data.Id);
                return dbClient.getString();
            }
        }
    }
}
