﻿using Butterfly.Messages.IOCodes;
using System;
using System.Text;

namespace Butterfly.Messages
{
    public class ServerMessage : IRecycable<ServerMessage>, ISendable, IServerNotificationMessage
    {
        private const int HEADER_SIZE = 4;
		private const int MAX_SIZE = 131072;
		private const int DEFAULT_SIZE = 512;
        private const int MAX_STRING_LENGTH = 500;
		private const int MAX_TTL_MS = 600000; //10 minutes

        private byte[] data;
        private int currentBodyLength;
        private int currentMaxSize; // Avoid function evaluation overhead for data.Length
        private int position;
        private IRecycleBin<ServerMessage> owner;
        private bool finishedPacket;
        private ISendSubject subject;
        private ushort mesID;
        private ushort peakID;
        private int packetDebugnumber;
        private ushort outputCode;

        public ServerMessage(int packetDebugNumber)
        {
            this.packetDebugnumber = packetDebugNumber;
			this.currentMaxSize = packetDebugnumber > -1 ? MAX_SIZE : DEFAULT_SIZE;
			this.data = new byte[currentMaxSize];
        }

        public ServerMessage(ushort outputCode)
        {
            // TODO: Complete member initialization
            this.outputCode = outputCode;
        }

        internal void Init(ushort messageID, ISendSubject subject)
        {
            //Console.WriteLine("Out -> " + messageID + "(" + (short)messageID + ") " + packetDebugnumber);
            this.mesID = messageID;
            this.finishedPacket = false;
            this.subject = subject;
            this.position = HEADER_SIZE;
            this.currentBodyLength = 0;
            this.AddInt16((short)messageID);

			//ResetLifetime();
        }

        internal void AddInt32(int value)
        {
			EnsureCapacity(4);

			this.data[position++] = (byte)(255 & (value >> 24));
			this.data[position++] = (byte)(255 & (value >> 16));
			this.data[position++] = (byte)(255 & (value >> 8));
			this.data[position++] = (byte)(255 & (value));
			this.currentBodyLength += 4;
        }

        internal void AddUInt32(uint value)
        {
			EnsureCapacity(4);

			this.data[position++] = (byte)(255 & (value >> 24));
			this.data[position++] = (byte)(255 & (value >> 16));
			this.data[position++] = (byte)(255 & (value >> 8));
			this.data[position++] = (byte)(255 & (value));
			this.currentBodyLength += 4;
        }

        internal void AddInt16(short value)
        {
			EnsureCapacity(2);

			this.data[position++] = (byte)(255 & (value >> 8));
			this.data[position++] = (byte)(255 & (value));
			this.currentBodyLength += 2;
        }

        internal void AddBoolean(bool value)
        {
            //this.AddInt32(value ? 1 : 0);
            this.AddByte((byte)(value ? 1 : 0));
        }

        internal void AddString(string value, bool allowMaxSize = false)
        {
			int length = value.Length;
			if (length > MAX_STRING_LENGTH && !allowMaxSize)
			{
				value = value.Substring(0, MAX_STRING_LENGTH);
				length = value.Length;
			}

			this.AddInt16((short)length);
			EnsureCapacity(length); //sizeof(char) = 2 * sizeof(byte)

            // TODO add UTF8 support (dario)
			for (int i = 0; i < length; i++)
			{
				this.data[position++] = (byte)(255 & value[i]);
			}

			this.currentBodyLength += length;
		}

		internal void AddChar(char value)
		{
			this.AddInt16(1);
			EnsureCapacity(1); 

			this.data[position++] = (byte)(255 & value);
			this.currentBodyLength++;
		}

		private void EnsureCapacity(int length)
		{
			if (finishedPacket)
				throw new InvalidOperationException("Packet already finalized!");

			if (this.position + 1 > MAX_SIZE)
				throw new IndexOutOfRangeException("Packet size overflow");

			if (this.position + length > this.currentMaxSize)
				IncreaseSize(this.position + length - this.currentMaxSize);
		}

        internal void AddByte(byte data)
        {
			EnsureCapacity(1);
            this.data[position++] = data;
            this.currentBodyLength++;
        }

        private void IncreaseSize(int bytesNeeded)
        {
            int newSize = (int)(this.currentMaxSize * 1.5f);
            if (newSize - this.currentMaxSize < bytesNeeded)
                newSize += bytesNeeded; //so large that it needs a lot of data apparently

            byte[] biggerArray = new byte[newSize];
            Array.Copy(data, biggerArray, this.position);
            this.currentMaxSize = newSize;
            this.data = biggerArray;
			this.peakID = this.mesID;
        }

        public byte[] Bytes()
        {
            if (finishedPacket)
                return data;

			this.data[0] = (byte)(255 & (this.currentBodyLength >> 24));
			this.data[1] = (byte)(255 & (this.currentBodyLength >> 16));
			this.data[2] = (byte)(255 & (this.currentBodyLength >> 8));
			this.data[3] = (byte)(255 & (this.currentBodyLength));

            finishedPacket = true;
            return data;
        }

        #region Debugging
        //internal ClientMessageUnitTest GetUnitTest()
        //{
        //    return new ClientMessageUnitTest(data.ToArray());
        //}
        #endregion

		public void SetOwner(IRecycleBin<ServerMessage> bin)
        {
            this.owner = bin;
			//this.watch = new Stopwatch();
			//ResetLifetime();
        }

		//private void ResetLifetime() // I know we all wish we had this in life from time to time
		//{
		//	if (watch != null)
		//		watch.Restart();
		//}

        public void Clean()
        {
            finishedPacket = false;
            this.position = HEADER_SIZE;
        }

        public void ReturnToOwner()
        {
            if (this.owner != null)
            {// AKA persistant packet
                this.owner.PutInBin(this);
            }
        }

        public int GetLength()
        {
            return this.position;
        }

        public void SendMe()
        {
            if (this.subject != null)
                this.SendMe(this.subject);
        }

        public void SendMe(ISendSubject aSuject)
        {
            aSuject.SendDataToConnection(Bytes(), GetLength());
        }

        internal string GetDebugMessage()
        {
            return this.mesID.ToString() + "[" + ((int)this.mesID) + "] - " + packetDebugnumber;
        }

        public int getID()
        {
            return this.packetDebugnumber;
        }

        public int GetMessageID()
        {
            return this.mesID;
        }

		internal void PrintStatus(StringBuilder builder)
		{
			builder.AppendLine();
			builder.AppendLine("Body length: " + this.currentBodyLength);
			builder.AppendLine("Current packet ID peak: " + this.peakID);
			//builder.AppendLine("Object life-time (ms): " + this.watch.ElapsedMilliseconds);
		}

		//internal bool TTLExpired()
		//{
		//	return (this.watch.ElapsedMilliseconds > MAX_TTL_MS);
		//}

		internal void Destroy()
		{
			//watch.Stop();
			//watch = null;
			data = null;
		}
    }
}