﻿using ConsoleWriter;
using Database_Manager.Database;
using Database_Manager.Session_Details.Interfaces;
using MySql.Data.MySqlClient;
using System;
using System.Data;
using System.Diagnostics;

namespace Database_Manager.Database.Session_Details
{
    internal class QueryAdapter : IRegularQueryAdapter
    {
        #region Fields
        protected MySqlClient client;
        protected MySqlCommand command;
        private Stopwatch watch;

        protected bool isExecutingAsync;
        private MySqlDataReader reader;
        private MySqlDataAdapter adapter;
        #endregion

        #region Constructor
        internal QueryAdapter(MySqlClient client)
        {
            this.client = client;
            this.watch = new Stopwatch();
            this.isExecutingAsync = false;
        }
        #endregion

        #region Methods
        #region General handling
        public void addParameter(string name, byte[] data)
        {
            this.command.Parameters.Add(new MySqlParameter(name, MySqlDbType.Blob, data.Length));
        }

        public void addParameter(string parameterName, object val)
        {
            this.command.Parameters.AddWithValue(parameterName, val);
        }

        public void setQuery(string query)
        {
            this.command.Parameters.Clear();
            this.command.CommandText = query;
        }
        #endregion

        #region Blocking/non async functions
        public bool findsResult()
        {
            watch.Restart();
            bool hasRows = false;
            try
            {
                using (reader = this.command.ExecuteReader())
                {
                    hasRows = reader.HasRows;
                }
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
            return hasRows;
        }

        public int getInteger()
        {
            watch.Restart();
            int result = 0;
            try
            {
                object obj2 = this.command.ExecuteScalar();
                if (obj2 != null)
                {
                    int.TryParse(obj2.ToString(), out result);
                }
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
            return result;
        }

        public DataRow getRow()
        {
            watch.Restart();
            DataRow row = null;
            try
            {
                DataTable dataTable = new DataTable();
                using (adapter = new MySqlDataAdapter(this.command))
                {
                    adapter.Fill(dataTable);
                }

                if (dataTable.Rows.Count > 0) //Found one or more results
                {
                    //Return the first one
                    row = dataTable.Rows[0];
                }
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
            return row;
        }

        public string getString()
        {
            watch.Restart();
            string str = string.Empty;
            try
            {
                object obj2 = this.command.ExecuteScalar();
                if (obj2 != null)
                {
                    str = obj2.ToString();
                }
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
            return str;
        }

        public DataTable getTable()
        {
            DataTable dataTable = new DataTable();
            watch.Restart();

            try
            {
                using (adapter = new MySqlDataAdapter(this.command))
                {
                    adapter.Fill(dataTable);
                }
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
            return dataTable;
        }

        public long insertQuery()
        {
            watch.Restart();
            long lastInsertedId = 0L;
            try
            {
                this.command.ExecuteScalar();
                lastInsertedId = this.command.LastInsertedId;
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
            return lastInsertedId;
        }

        public void runFastQuery(string query)
        {
            watch.Restart();
            this.setQuery(query);
            this.runQuery();
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
        }

        public void runQuery()
        {
            watch.Restart();
            try
            {
                this.command.ExecuteNonQuery();
            }
            catch (Exception exception)
            {
                Writer.LogQueryError(exception, this.command.CommandText);
            }
            DatabaseStats.totalQueryTime += watch.ElapsedMilliseconds;
            DatabaseStats.totalQueries++;
        }
        #endregion
        
        #endregion
    }
}

